(* Version is automatically generated from unicode data at
 * src/generator/data *)
val version : string

module Categories : sig

  val cc : Sedlex_cset.t
  val cf : Sedlex_cset.t
  val cn : Sedlex_cset.t
  val co : Sedlex_cset.t
  val cs : Sedlex_cset.t
  val ll : Sedlex_cset.t
  val lm : Sedlex_cset.t
  val lo : Sedlex_cset.t
  val lt : Sedlex_cset.t
  val lu : Sedlex_cset.t
  val mc : Sedlex_cset.t
  val me : Sedlex_cset.t
  val mn : Sedlex_cset.t
  val nd : Sedlex_cset.t
  val nl : Sedlex_cset.t
  val no : Sedlex_cset.t
  val pc : Sedlex_cset.t
  val pd : Sedlex_cset.t
  val pe : Sedlex_cset.t
  val pf : Sedlex_cset.t
  val pi : Sedlex_cset.t
  val po : Sedlex_cset.t
  val ps : Sedlex_cset.t
  val sc : Sedlex_cset.t
  val sk : Sedlex_cset.t
  val sm : Sedlex_cset.t
  val so : Sedlex_cset.t
  val zl : Sedlex_cset.t
  val zp : Sedlex_cset.t
  val zs : Sedlex_cset.t

  val list : (string * Sedlex_cset.t) list
end

module Properties : sig

  val alphabetic       : Sedlex_cset.t
  val ascii_hex_digit  : Sedlex_cset.t
  val hex_digit        : Sedlex_cset.t
  val id_continue      : Sedlex_cset.t
  val id_start         : Sedlex_cset.t
  val lowercase        : Sedlex_cset.t
  val math             : Sedlex_cset.t
  val other_alphabetic : Sedlex_cset.t
  val other_lowercase  : Sedlex_cset.t
  val other_math       : Sedlex_cset.t
  val other_uppercase  : Sedlex_cset.t
  val uppercase        : Sedlex_cset.t
  val white_space      : Sedlex_cset.t
  val xid_continue     : Sedlex_cset.t
  val xid_start        : Sedlex_cset.t

  val list : (string * Sedlex_cset.t) list
end
